
/*
 *  $Id: agPrint.c,v 2.1 1998/09/16 20:45:21 bkorb Exp $
 *
 *  The following code is necessary because the user can give us
 *  a printf format requiring a string pointer yet fail to provide
 *  a valid pointer, thus it will fault.  This code protects
 *  against the fault so an error message can be emitted instead of
 *  a core dump :-)
 */

/*
 *  Autogen copyright 1992-1998 Bruce Korb
 *
 *  Autogen is free software.
 *  You may redistribute it and/or modify it under the terms of the
 *  GNU General Public License, as published by the Free Software
 *  Foundation; either version 2, or (at your option) any later version.
 *
 *  Autogen is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Autogen.  See the file "COPYING".  If not,
 *  write to:  The Free Software Foundation, Inc.,
 *             59 Temple Place - Suite 330,
 *             Boston,  MA  02111-1307, USA.
 */

#include <signal.h>
#include <compat/compat.h>

#ifndef HAVE_STRSIGNAL
#  include <compat/strsignal.c>
#endif

#include "autogen.h"


STATIC sigjmp_buf printJumpEnv;

    STATIC void
printFault( int sig )
{
    siglongjmp( printJumpEnv, sig );
}

    int
safePrintf( char* pzBuf, size_t bufSize, char* pzFmt, va_list argV )
{
    tSCC zOvfloFmt[] = "%s ERROR:  format buffer overflow on\n\t`%s'\n";

    size_t printSize;
    int    faultType;
    struct sigaction  sa;
    struct sigaction  saSave1;
    struct sigaction  saSave2;

    sa.sa_handler = printFault;
    sa.sa_flags   = 0;
    sigemptyset( &sa.sa_mask );

    sigaction( SIGBUS,  &sa, &saSave1 );
    sigaction( SIGSEGV, &sa, &saSave2 );

    /*
     *  IF the vsprintf call below is going to address fault,
     *  THEN ...
     */
    if (faultType = sigsetjmp( printJumpEnv, 0 ),
        faultType != 0) {
        extern char* strsignal();

        /*
         *  IF the fprintf command in the then clause has not failed yet,
         *  THEN perform that fprintf
         */
        if (sigsetjmp( printJumpEnv, 0 ) == 0)
            fprintf( stderr, "%s ERROR:  %s processing printf format:\n"
                     "\t%s\n", autogenOptions.pzProgName,
                     strsignal( faultType ), pzFmt );

        /*
         *  The "vsprintf" command below faulted, so we exit
         */
        LOAD_ABORT;
    }

#ifndef PRINTF_RETURNS_VOID
    printSize = vsprintf( pzBuf, pzFmt, argV );

    if (printSize >= bufSize) {
        fprintf( stderr, zOvfloFmt, pzProg, pzFmt );
        LOAD_ABORT;
    }

    sigaction( SIGBUS,  &saSave1, (struct sigaction*)NULL );
    sigaction( SIGSEGV, &saSave2, (struct sigaction*)NULL );
    return (int)printSize;

#else
    {
        STATIC char* getSection( char* pzSection );
        int     ct              = 0;
        char    *pzSection      = getSection( pzFmt );
        char    *pz             = pzBuf;

        do  {
            sprintf( pz, pzSection, argV[ct++] );
            printSize = ( strlen( pz ) );
            if (strlen( pzBuf ) >= bufSize ) {
                fprintf( stderr, zOvfloFmt, pzProg, pzFmt );
                LOAD_ABORT;
            }
            pz += printSize;
            free (pzSection);

        } while ((pzSection = getSection( (char*)NULL )));
    }

    sigaction( SIGBUS,  &saSave1, (struct sigaction*)NULL );
    sigaction( SIGSEGV, &saSave2, (struct sigaction*)NULL );
    return (int)printSize;
#endif
}

#ifdef PRINTF_RETURNS_VOID

/*
 * Break a string up into tokens like strtok -- only we return copies
 * of the sections (which need to be free()ed), and the delimiter is any
 * valid printf format specifier sequence which *is* returned as part
 * of the token (if it exists).  Returns NULL when no more tokens are
 * available.
 */
    STATIC char*
getSection( char *pzSection )
{
    static char* pzScan = (char*)NULL;
    char*        pzStart;
    size_t       len;

    /*
     * if pzSection is supplied, sectionalize from the start
     */
    if (pzSection != (char*)NULL)
        pzScan = pzSection;

    /*
     * return if we have reached the end
     */
    if ((pzScan == (char*)NULL) || (*pzScan == NUL))
        return (char*)NULL;

    pzStart = pzScan;

    /*
     * find next format specifier
     */
    for (;;) {
        pzScan = strchr( pzScan, FMT_CHAR );

        if (pzScan == NULL) {
            len = strlen( pzStart );
            break;
        }

        if (pzScan[1] != FMT_CHAR) {
            /*
             *  Advance the scan pointer to the character following
             *  the valid format specifier character (or string end).
             */
            pzScan += strcspn( pzScan, VALID_FORMAT_SPECS);

            if (*pzScan == NUL) {
                len = pzScan - pzStart;
                pzScan = (char*)NULL;

            } else {
                pzScan++;
                len = pzScan - pzStart;
            }
            break;
        }

        /*
         *  Skip past the 'escaped' format character and keep looking
         */
        pzScan += 2;
    }

    /*
     * copy into new memory
     */
    {
        char* pzResult = (char*)AGALOC( 1 + len );

        if (pzResult == (char*)NULL) {
            fprintf( stderr, zAllocErr, autogenOptions.pzProgName,
                     len+1, "Shell Output" );
            LOAD_ABORT;
        }

        strncpy( pzResult, pzStart, len);
        pzResult[len]  = NUL;

        return pzResult;
    }
}
#endif /* PRINTF_RETURNS_VOID */
